using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMCCD - Custom Clothing Decals v1.0
//
// Plugin allowing customization of the decal channel in characters outfits

namespace VAMCCDPlugin
{
    public class VAMCCD : MVRScript
    {
	    private static string PLUGIN_PATH;
	    private static string SHADERS_BUNDLE_PATH = "/vamccd_assets.shaderbundle";
	    private static string SHADER_BASE_PATH = "Assets/VAMCCD/vamccd-customblitcopy.shader";
	    private static string MATERIAL_BASE_PATH = "Assets/VAMCCD/vamccd_defaultmat.mat";
	    private static string BUNDLED_DECALS_PATH = "assets/vamccd/bundled_decals/";
	    private static List<string> bundleDecalsFiles = new List<string>() {"None"};
	    private static List<string> textureWrapModes = new List<string>() {"Repeat","Repeat Mirror","No Repeat","Mirror Once"};
	    private static List<string> filterTexture6 = new List<string>() {"_DetailTex", "_DetailMap"}; // Some clothes from meshed have a different channel for the 6th texture
	    
	    private string lastFileExplorerPath = "Custom/Clothing";

	    private DAZCharacter currentChara;
		private DAZCharacterSelector myGeometry;
		private DAZClothingItem[] currentClothings;
		private DAZMergedSkinV2 _dazSkin;

		// Shader & stuff
		private Shader blitShader;
		private Material blitMaterial;
		
		// Dictionaries (resources)
		private Dictionary<string, Material> decalsMaterials = new Dictionary<string, Material>();
		private Dictionary<string, Texture2D> decalsSourceTextures = new Dictionary<string, Texture2D>();
		private Dictionary<string, Texture2D> decalsDestTextures = new Dictionary<string, Texture2D>();
		
		// Dictionaries (save data)
		private Dictionary<string, JSONStorableColor> decalsColors = new Dictionary<string, JSONStorableColor>();
		private Dictionary<string, JSONStorableUrl> decalsUrls = new Dictionary<string, JSONStorableUrl>();
		private Dictionary<string, JSONStorableStringChooser> decalsFileBundled = new Dictionary<string, JSONStorableStringChooser>();
		private Dictionary<string, JSONStorableStringChooser> decalsWrapMode = new Dictionary<string, JSONStorableStringChooser>();
		private Dictionary<string, JSONStorableFloat> decalsRotation = new Dictionary<string, JSONStorableFloat>();
		private Dictionary<string, JSONStorableFloat> decalsTwirlStrength = new Dictionary<string, JSONStorableFloat>();
		private Dictionary<string, JSONStorableFloat> decalsTwirlOffsetX = new Dictionary<string, JSONStorableFloat>();
		private Dictionary<string, JSONStorableFloat> decalsTwirlOffsetY = new Dictionary<string, JSONStorableFloat>();
		
		// Dictionaries (clothing related)
		private Dictionary<string, JSONStorableFloat> decalsTileX = new Dictionary<string, JSONStorableFloat>();
		private Dictionary<string, JSONStorableFloat> decalsTileY = new Dictionary<string, JSONStorableFloat>();
		
		// Dictionaries (UI)
		private Dictionary<string, List<UIDynamic>> decalsUI = new Dictionary<string, List<UIDynamic>>();
		
		// "ALL" UI elements (to clear the UI)
		private List<UIDynamicColorPicker> allUIColorPickers;
		private List<UIDynamicSlider> allUISliders;
		private List<UIDynamicToggle> allUIToggles;
		private List<UIDynamicTextField> allUITextFields;
		private List<UIDynamicPopup> allUIPopups;
		private List<UIDynamicButton> allUIButtons;
		private List<UIDynamic> allUISpacers;
		
		// Assetbundle
		private Request bundleRequest = null;
		
		// Plugin data (due to the fact that the UI is instantiated after init, I need to grab my data and restore manually
		public JSONClass decalsDatas { get; set; }
		
		// Init
		private bool isBundleLoaded = false;
		private bool isInit = false;
		
		// Clothing set checks
		private float internalTimer = 0f;
		private float clothingSetCheckRate = 3f;
		private float clothingSetNextCheck = 0f;
		private int clothingSetCurrentCount = 0;

		// The only global UI components
		private JSONStorableStringChooser ClothingSelector;
		protected List<string> clothingChoices;
		private JSONStorableString DecalsUIInfos;
		private UIDynamic DecalsUIInfosField;
		
		
		// Some default texts
		public static string TxtNoClothes = "<color=#000><size=35><b>Woops!</b></size></color>\n\n<size=30><color=#333>It seems your character is naked. There are no clothes to add decals on.\n\n(or maybe the clothings you've chosen are not compatible with decals... bummer)</color></size>";

		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "Person")
                {
                    SuperController.LogError("Please add VAMCCD on a Person");
                    return;
                }
				
                PLUGIN_PATH = GetPluginPath(this);

                DecalsUIInfos = new JSONStorableString("Decals Infos", TxtNoClothes) { isStorable=false,isRestorable=false };
                DecalsUIInfosField = CreateTextField(DecalsUIInfos, false);
                DecalsUIInfosField.height = 350.0f;
                ToggleUIDynamic(false, DecalsUIInfosField);
                
                JSONStorableString _helpText = new JSONStorableString("Help",
	                "<color=#000><size=35><b>VAMCCD:\nCustom Clothing Decals</b></size></color>\n\n" + 
	                "<size=30><color=#333>" +
	                "This plugin allows you customize decals of all clothes active on your character.\n\n" +
	                "- Select one of the clothing item in the dropdown.\n" +
	                "- Edit the settings to your liking\n\n" +
	                "<b>Notes:</b>\n" +
	                "- A handful of decals are available by default, but you can select custom decals of your own.\n" +
	                "- Some clothes are not compatible with decals. They will not appear in the plugin.\n" +
	                "- The wrap mode (the way the texture repeats) might be hard to understand. This is documented on the hub." +
	                "</color></size>"
                );
				
                UIDynamic HTUID = CreateTextField(_helpText, true);
                HTUID.height = 950.0f;
                
				blitShader = Shader.Find("Hidden/BlitCopy");
				blitMaterial = new Material(blitShader);
				
				myGeometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;

				// Creating all the lists 
				allUIColorPickers = new List<UIDynamicColorPicker>();
				allUISliders = new List<UIDynamicSlider>();
				allUIToggles = new List<UIDynamicToggle>();
				allUIPopups = new List<UIDynamicPopup>();
				allUITextFields = new List<UIDynamicTextField>();
				allUIButtons = new List<UIDynamicButton>();
				allUISpacers = new List<UIDynamic>();
				
				// The dropdown to edit the clothes
				clothingChoices = new List<string>();
				ClothingSelector = new JSONStorableStringChooser("Clothing", clothingChoices, "", "Clothing") {isStorable=false,isRestorable=false};
				ClothingSelector.setCallbackFunction += (val) => { ToggleClothingId(val); };
				UIDynamicPopup CSdp = CreateScrollablePopup(ClothingSelector, false);
				CSdp.labelWidth = 120f;
				
				// Loading the assetbundle containing the fluids
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + SHADERS_BUNDLE_PATH, callback = OnBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
            }
            catch(Exception e)
            {
                SuperController.LogError("VAMCCD - Exception caught: " + e);
            }
		}


		void Start()
        {
			try
            {

			}
            catch(Exception e)
            {
                SuperController.LogError("VAMCCD - Exception caught: " + e);
            }
        }
		
		void Update() {
			if (isInit == false && !SuperController.singleton.isLoading && isBundleLoaded == true)
			{
				isInit = true;
				clothingSetNextCheck = internalTimer + clothingSetCheckRate;
				InitDecalsManager();
			}

			if (isInit == true)
			{
				internalTimer += Time.deltaTime;
				if (internalTimer > clothingSetNextCheck)
				{
					clothingSetNextCheck = internalTimer + clothingSetCheckRate;
					if ( hasClothingSetChanged() )
					{
						// Getting the current data to restore properly the new fields that are gonna be generated
						// (if some of the same clothes are kept)
						decalsDatas = GetJSON();
						// Clearing UI then creating the new one
						ClearUI();
						InitDecalsManager();
					}
				}
			}
		}
		
		// **************************
		// Functions
		// **************************
		// I'm working with a unique decal id (the storable value) for every item, if this one exists, the others too
		bool decalUniqueIdExists( string decalUniqueId )
		{
			return decalsSourceTextures.ContainsKey(decalUniqueId);
		}

		void InitDecalsManager()
		{
			try
			{
				HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
				
				// Creating JSONStorables
				UIDynamicTextField tmpText;
				UIDynamicPopup tmpPopup;
	            UIDynamicSlider tmpSlider;
	            UIDynamicColorPicker tmpColor;
	            UIDynamicButton tmpButton;
	            UIDynamicToggle tmpToggle;
	            UIDynamic tmpSpacer;
	            float tmpFloat;

	            currentClothings = myGeometry.clothingItems;
				clothingSetCurrentCount = 0;
				clothingChoices = new List<string>();
				foreach( DAZClothingItem clItem in currentClothings ) {
					if( clItem.active )
					{
						clothingSetCurrentCount++;
						DAZSkinWrap[] skinWraps = clItem.GetComponentsInChildren<DAZSkinWrap>();

						foreach (DAZSkinWrap swItem in skinWraps)
						{
							// It seems that Meshed uses the same material settings (and makes the UI not super user friendly) for all wraps
							// So we only need the "normal" wrap, it should update the material properly everytime
							if (swItem.wrapName != "Normal") continue;
							
							DAZSkinWrapMaterialOptions[] skinWrapsOpts = swItem.transform.gameObject.GetComponents<DAZSkinWrapMaterialOptions>();
							
							foreach (DAZSkinWrapMaterialOptions swMo in skinWrapsOpts)
							{
								// If we got a label set, then it's different than the usual "Decal" slot, which means, decal is not used, so we can skip this
								if (swMo.customTexture6Label != null)
								{
									if (swMo.customTexture6Label.text != "" && swMo.customTexture6Label.text != null && filterTexture6.Contains(swMo.customTexture6Label.text )) continue;
								}
								
								/* ********************** UI INIT START ************************** */
								/* Yes I know, this is a frikking "unoptimized" way to produce the UI... but whatever :D
								 * Someday I'll write a UI helper haha
								 */
								string decalUniqueId = swMo.storeId;

								clothingChoices.Add(decalUniqueId);
								var clItemUI = new List<UIDynamic>();
								
								// -- COLOR
								JSONStorableColor DecalColor = new JSONStorableColor("Color " + decalUniqueId, hsvcDefault);
								DecalColor.setJSONCallbackFunction += (val) => { ColorCallback(val,swMo.storeId); };
								tmpColor = CreateColorPicker(DecalColor);
								tmpColor.label = "Color";
								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsColors.Add(decalUniqueId, DecalColor);
								}
								else
								{
									DeregisterColor(decalsColors[decalUniqueId]);
									decalsColors[decalUniqueId] = DecalColor;
								}

								allUIColorPickers.Add(tmpColor);
								clItemUI.Add((UIDynamic)tmpColor);
								RegisterColor(DecalColor);

								// -- BUNDLED FILE
								JSONStorableStringChooser DecalFile = new JSONStorableStringChooser("Bundled Decal " + decalUniqueId, bundleDecalsFiles, "None", "Bundled Decal");
								DecalFile.setCallbackFunction += (val) => { SetBundledDecalCallback(swMo.storeId); };
								tmpPopup = CreateScrollablePopup(DecalFile, false);
								tmpPopup.labelWidth = 120f;
								allUIPopups.Add(tmpPopup);
								clItemUI.Add((UIDynamic)tmpPopup);
								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsFileBundled.Add(decalUniqueId, DecalFile);
								}
								else
								{
									DeregisterStringChooser(decalsFileBundled[decalUniqueId]);
									decalsFileBundled[decalUniqueId] = DecalFile;
								}
								RegisterStringChooser(DecalFile);
								
								// -- URL BTN
								tmpButton = CreateButton("Select a custom decal", false);
								allUIButtons.Add(tmpButton);
								clItemUI.Add((UIDynamic)tmpButton);
								tmpButton.button.onClick.AddListener( () => { SetCustomDecalCallback(swMo.storeId); } );

								// -- URL
								JSONStorableUrl DecalUrl = new JSONStorableUrl("DecalUrl " + decalUniqueId, "");
								tmpText = CreateTextField(DecalUrl);
								setupTextField(tmpText, 80f);
								allUITextFields.Add(tmpText);
								clItemUI.Add((UIDynamic)tmpText);
								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsUrls.Add(decalUniqueId, DecalUrl);
								}
								else
								{
									DeregisterUrl(decalsUrls[decalUniqueId]);
									decalsUrls[decalUniqueId] = DecalUrl;
								}
								RegisterUrl(DecalUrl);
								
								// -- WRAP MODE
								JSONStorableStringChooser DecalWrapMode = new JSONStorableStringChooser("Wrap Mode " + decalUniqueId, textureWrapModes, "Repeat", "Wrap Mode");
								DecalWrapMode.setCallbackFunction += (val) => { renderDecalId(swMo.storeId); };
								tmpPopup = CreateScrollablePopup(DecalWrapMode, false);
								tmpPopup.labelWidth = 120f;
								allUIPopups.Add(tmpPopup);
								clItemUI.Add((UIDynamic)tmpPopup);
								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsWrapMode.Add(decalUniqueId, DecalWrapMode);
								}
								else
								{
									DeregisterStringChooser(decalsWrapMode[decalUniqueId]);
									decalsWrapMode[decalUniqueId] = DecalWrapMode;
								}
								RegisterStringChooser(DecalWrapMode);
								
								// -- ROTATION
								JSONStorableFloat DecalRotation = new JSONStorableFloat("Rotation " + decalUniqueId, 0f, 0, 360f);
								tmpSlider = CreateSlider(DecalRotation);
								tmpSlider.valueFormat = "F2";
								tmpSlider.labelText.text = "Rotation";
								allUISliders.Add(tmpSlider);
								clItemUI.Add((UIDynamic)tmpSlider);
								DecalRotation.setCallbackFunction += (val) => { renderDecalId(swMo.storeId); };
								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsRotation.Add(decalUniqueId, DecalRotation);
								}
								else
								{
									DeregisterFloat(decalsRotation[decalUniqueId]);
									decalsRotation[decalUniqueId] = DecalRotation;
								}
								RegisterFloat(DecalRotation);
								
								// -- TWIRL STRENGTH
								JSONStorableFloat DecalTwirlStrength = new JSONStorableFloat("Twirl Strength " + decalUniqueId, 0f, -20f, 20f);
								tmpSlider = CreateSlider(DecalTwirlStrength);
								tmpSlider.valueFormat = "F2";
								tmpSlider.labelText.text = "Twirl Strength";
								allUISliders.Add(tmpSlider);
								clItemUI.Add((UIDynamic)tmpSlider);
								DecalTwirlStrength.setCallbackFunction += (val) => { renderDecalId(swMo.storeId); };
								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsTwirlStrength.Add(decalUniqueId, DecalTwirlStrength);
								}
								else
								{
									DeregisterFloat(decalsTwirlStrength[decalUniqueId]);
									decalsTwirlStrength[decalUniqueId] = DecalTwirlStrength;
								}
								RegisterFloat(DecalTwirlStrength);
								
								// -- TWIRL OFFSET X
								JSONStorableFloat DecalTwirlOffsetX = new JSONStorableFloat("Twirl Offset X " + decalUniqueId, 0.5f, 0, 1f);
								tmpSlider = CreateSlider(DecalTwirlOffsetX);
								tmpSlider.valueFormat = "F2";
								tmpSlider.labelText.text = "Twirl Offset X";
								allUISliders.Add(tmpSlider);
								clItemUI.Add((UIDynamic)tmpSlider);
								DecalTwirlOffsetX.setCallbackFunction += (val) => { renderDecalId(swMo.storeId); };
								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsTwirlOffsetX.Add(decalUniqueId, DecalTwirlOffsetX);
								}
								else
								{
									DeregisterFloat(decalsTwirlOffsetX[decalUniqueId]);
									decalsTwirlOffsetX[decalUniqueId] = DecalTwirlOffsetX;
								}
								RegisterFloat(DecalTwirlOffsetX);
								
								// -- TWIRL OFFSET Y
								JSONStorableFloat DecalTwirlOffsetY = new JSONStorableFloat("Twirl Offset Y " + decalUniqueId, 0.5f, 0f, 1f);
								tmpSlider = CreateSlider(DecalTwirlOffsetY);
								tmpSlider.valueFormat = "F2";
								tmpSlider.labelText.text = "Twirl Offset Y";
								allUISliders.Add(tmpSlider);
								clItemUI.Add((UIDynamic)tmpSlider);
								DecalTwirlOffsetY.setCallbackFunction += (val) => { renderDecalId(swMo.storeId); };
								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsTwirlOffsetY.Add(decalUniqueId, DecalTwirlOffsetY);
								}
								else
								{
									DeregisterFloat(decalsTwirlOffsetY[decalUniqueId]);
									decalsTwirlOffsetY[decalUniqueId] = DecalTwirlOffsetY;
								}
								RegisterFloat(DecalTwirlOffsetY);
								
								// -- CLOTHING CONTROLS
								tmpFloat = containingAtom.GetStorableByID(decalUniqueId).GetFloatParamValue("customTexture6TileX");
								JSONStorableFloat DecalTileX = new JSONStorableFloat("Clothing Tile X", 1f, 0f, 10f, false);
								DecalTileX.val = tmpFloat;
								tmpSlider = CreateSlider(DecalTileX);
								tmpSlider.valueFormat = "F3";
								allUISliders.Add(tmpSlider);
								clItemUI.Add((UIDynamic)tmpSlider);
								DecalTileX.setCallbackFunction += (val) => { TileXCallback(DecalTileX, swMo.storeId); };
								
								tmpFloat = containingAtom.GetStorableByID(decalUniqueId).GetFloatParamValue("customTexture6TileY");
								JSONStorableFloat DecalTileY = new JSONStorableFloat("Clothing Tile Y", 1f, 0f, 10f, false);
								DecalTileY.val = tmpFloat;
								tmpSlider = CreateSlider(DecalTileY);
								tmpSlider.valueFormat = "F3";
								allUISliders.Add(tmpSlider);
								clItemUI.Add((UIDynamic)tmpSlider);
								DecalTileY.setCallbackFunction += (val) => { TileYCallback(DecalTileY, swMo.storeId ); };
								
								tmpFloat = containingAtom.GetStorableByID(decalUniqueId).GetFloatParamValue("customTexture6OffsetX");
								JSONStorableFloat DecalOffsetX = new JSONStorableFloat("Clothing Offset X", 0f, -1f, 1f, false);
								DecalOffsetX.val = tmpFloat;
								tmpSlider = CreateSlider(DecalOffsetX);
								tmpSlider.valueFormat = "F3";
								allUISliders.Add(tmpSlider);
								clItemUI.Add((UIDynamic)tmpSlider);
								DecalOffsetX.setCallbackFunction += (val) => { OffsetXCallback(DecalOffsetX, swMo.storeId); };
								
								tmpFloat = containingAtom.GetStorableByID(decalUniqueId).GetFloatParamValue("customTexture6OffsetY");
								JSONStorableFloat DecalOffsetY = new JSONStorableFloat("Clothing Offset Y", 0f, -1f, 1f, false);
								DecalOffsetY.val = tmpFloat;
								tmpSlider = CreateSlider(DecalOffsetY);
								tmpSlider.valueFormat = "F3";
								allUISliders.Add(tmpSlider);
								clItemUI.Add((UIDynamic)tmpSlider);
								DecalOffsetY.setCallbackFunction += (val) => { OffsetYCallback(DecalOffsetY, swMo.storeId ); };

								tmpSpacer = CreateSpacer();
								allUISpacers.Add(tmpSpacer);
								clItemUI.Add( tmpSpacer);
								

								if (!decalUniqueIdExists(decalUniqueId))
								{
									decalsUI.Add(decalUniqueId, clItemUI);
								}
								else
								{
									decalsUI[decalUniqueId] = clItemUI;
								}

								ToggleDecalEditUIId(false,decalUniqueId);
								/* ********************** UI INIT END ************************** */
								
								/* ********************** DATA LOADING START ************************** */
								// Loading the data from the scene save
								// It will only happen on restore, if not, the plugin use defaults settings
								if (decalsDatas != null)
								{
									DecalColor.RestoreFromJSON( decalsDatas );
									DecalFile.RestoreFromJSON( decalsDatas );
									DecalUrl.RestoreFromJSON( decalsDatas );
									DecalWrapMode.RestoreFromJSON( decalsDatas );
									DecalRotation.RestoreFromJSON( decalsDatas );
									DecalTwirlStrength.RestoreFromJSON( decalsDatas );
									DecalTwirlOffsetX.RestoreFromJSON( decalsDatas );
									DecalTwirlOffsetY.RestoreFromJSON( decalsDatas );
								}
								/* ********************** DATA LOADING END ************************** */
								
								/* ********************** RESOURCES INIT START ************************** */
								// Due to the fact that there is absolutely no way to handle the async load of the texture by default
								// I'm emptying the decal slot to ensure nothing is saved in the scene (which would cause an override at load)
								// All the textures are gonna be loaded from my datas
								if( containingAtom.GetStorableByID(decalUniqueId) != null ) {
									if (containingAtom.GetStorableByID(decalUniqueId).GetUrlParamValue("customTexture_DecalTex") != "")
									{
										containingAtom.GetStorableByID(decalUniqueId).SetUrlParamValue("customTexture_DecalTex", "");
									}
								}

								if( !decalUniqueIdExists(decalUniqueId) ) {
									decalsSourceTextures.Add(decalUniqueId,getDecalTexture( decalUniqueId ) );
									decalsDestTextures.Add(decalUniqueId, new Texture2D( decalsSourceTextures[decalUniqueId].width, decalsSourceTextures[decalUniqueId].height, TextureFormat.RGBA32, false, true));
									foreach (int mtlSlot in swMo.paramMaterialSlots)
									{
										decalsMaterials.Add(decalUniqueId,swItem.GPUmaterials[mtlSlot]);
										break;
									}
									
								}
								else
								{
									// I'm gonna update the decal to be sure, maybe the user has changed it in the clothes
									// The rest is not changing
									decalsSourceTextures[decalUniqueId] = getDecalTexture( decalUniqueId );
								}
								/* ********************** RESOURCES INIT END ************************** */

								
								// Inits!
								ColorCallback(DecalColor,decalUniqueId);
							}
						}
					}
				}
				
				// After all that, updating the dropdown to select clothings
				ClothingSelector.val = ""; // This is important to produce a callback if we were already on the same value for the selected item
				if (clothingSetCurrentCount > 0)
				{
					ClothingSelector.choices = clothingChoices;
					ClothingSelector.val = clothingChoices.First();
					ToggleUIDynamic(false, DecalsUIInfosField);
				}
				else
				{
					ClothingSelector.choices = new List<string>();
					ToggleUIDynamic(true, DecalsUIInfosField);
				}
			}
			catch(Exception e)
			{
				SuperController.LogError("VAMCCD - Exception caught: " + e);
			}
		}
		
		// Thanks to Chokaphi (Decal Maker)
		// I had a hardtime fixing a memory leak in my initial code
		private void GpuCombine(string decalUniqueId, Material material)
		{
			int w = decalsSourceTextures[decalUniqueId].width;
			int h = decalsSourceTextures[decalUniqueId].height;
			
			RenderTextureReadWrite renderTexConversion = RenderTextureReadWrite.sRGB;
			
			RenderTexture tmp = RenderTexture.GetTemporary(w, h, 0, RenderTextureFormat.ARGB32, renderTexConversion);        //tmp.antiAliasing = 8;                                                                                                                           //Backup the currently set RenderTexture
			RenderTexture previous = RenderTexture.active;
			Graphics.Blit(decalsSourceTextures[decalUniqueId], tmp, material);
			RenderTexture.active = tmp;
			Graphics.CopyTexture(tmp, decalsDestTextures[decalUniqueId]);
			// Reset the active RenderTexture
			RenderTexture.ReleaseTemporary(tmp);
			// Applying Wrap mode
			decalsDestTextures[decalUniqueId].wrapMode = getDecalWrapModeEnum(decalsWrapMode[decalUniqueId].val);
		}

		private void updateDecalId(string decalUniqueId)
		{
			if (!decalUniqueIdExists(decalUniqueId)) return;
			
			decalsSourceTextures[decalUniqueId] = getDecalTexture(decalUniqueId);
			if (decalsSourceTextures[decalUniqueId].width != decalsDestTextures[decalUniqueId].width || decalsSourceTextures[decalUniqueId].height != decalsDestTextures[decalUniqueId].height)
			{
				decalsDestTextures[decalUniqueId] = new Texture2D(decalsSourceTextures[decalUniqueId].width, decalsSourceTextures[decalUniqueId].height, TextureFormat.RGBA32, false, true);
			}

			renderDecalId(decalUniqueId);
		}
		
		private void renderDecalId(string decalUniqueId)
		{
			if (!decalUniqueIdExists(decalUniqueId)) return;
			var targetColor = decalsColors[decalUniqueId];
			blitMaterial.SetTexture("_MainTex", decalsSourceTextures[decalUniqueId]);
			blitMaterial.SetFloat("_Rotation1", decalsRotation[decalUniqueId].val);
			blitMaterial.SetFloat("_TwirlStrength", decalsTwirlStrength[decalUniqueId].val);
			blitMaterial.SetVector("_TwirlCenter", new Vector4( decalsTwirlOffsetX[decalUniqueId].val, decalsTwirlOffsetY[decalUniqueId].val,0f,0f) );
			blitMaterial.SetColor("_Color", new Color( targetColor.colorPicker.currentColor.r, targetColor.colorPicker.currentColor.g, targetColor.colorPicker.currentColor.b, 1f));
			GpuCombine(decalUniqueId, blitMaterial);
			decalsMaterials[ decalUniqueId ].SetTexture("_DecalTex", decalsDestTextures[decalUniqueId]);
		}

		private bool hasClothingSetChanged()
		{
			currentClothings = myGeometry.clothingItems;
			int clothingCount = 0;
			foreach (DAZClothingItem clItem in currentClothings)
			{
				if (clItem.active)
				{
					clothingCount++;
				}
			}

			if (clothingCount != clothingSetCurrentCount) return true;

			return false;
		}

		private Texture2D getDecalTexture( string decalUniqueId )
		{
			if (decalsUrls[decalUniqueId].val != "" && decalsUrls[decalUniqueId].val != null)
			{
				return GetResource(decalsUrls[decalUniqueId].val);
			}
			else
			{
				// We want a bundled decal
				if (decalsFileBundled[decalUniqueId].val != "None")
				{
					return bundleRequest.assetBundle.LoadAsset<Texture2D>(BUNDLED_DECALS_PATH + decalsFileBundled[decalUniqueId].val + ".png");
				}
				// Or we want to clear (GetResource will load a transparent decal)
				else
				{
					return GetResource("");
				}
			}
		}

		private TextureWrapMode getDecalWrapModeEnum(string value)
		{
			switch (value)
			{
				case "No Repeat":
					return TextureWrapMode.Clamp;
				case "Repeat Mirror":
					return TextureWrapMode.Mirror;
				case "Mirror Once":
					return TextureWrapMode.MirrorOnce;
				default:
					return TextureWrapMode.Repeat;
			}
		}

		// Function clearing the whole CUA UI
		private void ClearUI()
		{		
			foreach( UIDynamicColorPicker uiEl in allUIColorPickers ) {
				RemoveColorPicker(uiEl);
			}
			allUIColorPickers.Clear();
			
			foreach( UIDynamicSlider uiEl in allUISliders ) {
				RemoveSlider(uiEl);
			}
			allUISliders.Clear();
			
			foreach( UIDynamicToggle uiEl in allUIToggles ) {
				RemoveToggle(uiEl);
			}
			allUIToggles.Clear();
			
			foreach( UIDynamicTextField uiEl in allUITextFields ) {
				RemoveTextField(uiEl);
			}
			allUITextFields.Clear();

			foreach (UIDynamicPopup uiEl in allUIPopups)
			{
				RemovePopup(uiEl);
			}
			
			foreach( UIDynamic uiEl in allUISpacers ) {
				RemoveSpacer(uiEl);
			}
			allUISpacers.Clear();
			
			foreach( UIDynamicButton uiEl in allUIButtons ) {
				RemoveButton(uiEl);
			}
			allUIButtons.Clear();
		}
		
		private void setupTextField( UIDynamicTextField target, float fieldHeight ) {
			target.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement tfLayout = target.GetComponent<LayoutElement>();
			tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
			target.height = fieldHeight;
			disableScrollOnText(target);
		}
		
		private void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}

		// **************************
		// Callbacks & similar
		// **************************
		private void OnBundleLoaded(Request aRequest) {
			try
			{
				bundleRequest = aRequest;
				Shader customShader = bundleRequest.assetBundle.LoadAsset<Shader>(SHADER_BASE_PATH);
				blitMaterial = new Material(customShader);
				
				// Grabbing all bundled pngs
				foreach (string bundleFileItem in bundleRequest.assetBundle.GetAllAssetNames())
				{
					if (bundleFileItem.Contains(BUNDLED_DECALS_PATH))
					{
						bundleDecalsFiles.Add(bundleFileItem.Replace(BUNDLED_DECALS_PATH, "").Replace(".png", ""));
					}
				}
				
				isBundleLoaded = true;
			}
			catch(Exception e)
			{
				SuperController.LogError("VAMCCD - Exception caught: " + e);
			}
		}

		private void SetBundledDecalCallback(string decalUniqueId)
		{
			decalsUrls[decalUniqueId].val = "";
			updateDecalId(decalUniqueId);
		}

		private void SetCustomDecalCallback(string decalUniqueId)
		{
			SuperController.singleton.fileBrowserUI.defaultPath = lastFileExplorerPath;
			SuperController.singleton.fileBrowserUI.SetTitle("Select a texture");
			SuperController.singleton.fileBrowserUI.fileFormat = "png|jpg|tga|tiff";
			SuperController.singleton.fileBrowserUI.showFiles = true;
			SuperController.singleton.fileBrowserUI.showDirs = true;
			SuperController.singleton.fileBrowserUI.SetTextEntry(false);
			SuperController.singleton.fileBrowserUI.hideExtension = false;
			SuperController.singleton.fileBrowserUI.showInstallFolderInDirectoryList = false;
			List<ShortCut> shortCutList = new List<ShortCut>();
			ShortCut shortAp = new ShortCut();
			shortAp.package = "AddonPackages Filtered";
			shortAp.path = "AddonPackages";
			shortAp.displayName = "Filter: Custom";
			shortAp.packageFilter = "Custom";
			shortCutList.Add(shortAp);
			
			ShortCut shortCutCustom = new ShortCut();
			shortCutCustom.path = "Custom\\";
			shortCutCustom.displayName = "Custom";
			shortCutCustom.flatten = false;
			shortCutList.Add(shortCutCustom);

			ShortCut shortCutBase = new ShortCut();
			shortCutBase.path = "Custom\\Clothing\\";
			shortCutBase.displayName = "Custom/Clothing";
			shortCutBase.flatten = false;
			shortCutList.Add(shortCutBase);

			if (FileManagerSecure.DirectoryExists("Custom/Textures"))
			{
				ShortCut shortCutTex = new ShortCut();
				shortCutTex.path = "Custom\\Textures\\";
				shortCutTex.displayName = "Custom/Textures";
				shortCutTex.flatten = false;
				shortCutList.Add(shortCutTex);
			}

			SuperController.singleton.fileBrowserUI.shortCuts = shortCutList;			
			SuperController.singleton.fileBrowserUI.Show((val) =>
			{
				if (val != "")
				{
					lastFileExplorerPath = val.Substring(0, val.LastIndexOfAny(new char[] {'/', '\\'}));
					decalsFileBundled[decalUniqueId].val = "None";
					decalsUrls[decalUniqueId].val = val;
					updateDecalId(decalUniqueId);
				}
			});
		}
		
		private void TileXCallback(JSONStorableFloat target, string decalUniqueId)
		{
			containingAtom.GetStorableByID(decalUniqueId).SetFloatParamValue("customTexture6TileX", target.val);
		}
		
		private void TileYCallback(JSONStorableFloat target, string decalUniqueId)
		{
			containingAtom.GetStorableByID(decalUniqueId).SetFloatParamValue("customTexture6TileY", target.val);
		}
		
		private void OffsetXCallback(JSONStorableFloat target, string decalUniqueId)
		{
			containingAtom.GetStorableByID(decalUniqueId).SetFloatParamValue("customTexture6OffsetX", target.val);
		}
		
		private void OffsetYCallback(JSONStorableFloat target, string decalUniqueId)
		{
			containingAtom.GetStorableByID(decalUniqueId).SetFloatParamValue("customTexture6OffsetY", target.val);
		}

		private void ColorCallback(JSONStorableColor targetColorstring, string decalUniqueId)
		{
			if (!decalUniqueIdExists(decalUniqueId)) return;
			renderDecalId(decalUniqueId);
		}

		protected void ToggleClothingId(string decalUniqueId)
		{
			foreach (var clUI in decalsUI)
			{
				bool showClothingUI = false;
				if (clUI.Key == decalUniqueId) showClothingUI = true;
				ToggleDecalEditUIId(showClothingUI, clUI.Key);
			}
		}
		protected void ToggleDecalEditUIId(bool enabled, string decalUniqueId)
		{
			if (!decalsUI.ContainsKey(decalUniqueId)) return;
			foreach( UIDynamic uidElem in decalsUI[decalUniqueId] )
			{
				ToggleUIDynamic(enabled, uidElem);
			}
		}

		protected void ToggleUIDynamic(bool enabled, UIDynamic target)
		{
			if (target == null) return;
			
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();
			if (tmpLE != null)
			{
				if (enabled == true)
				{
					tmpLE.transform.localScale = new Vector3(1, 1, 1);
					tmpLE.ignoreLayout = false;
				}
				else
				{
					tmpLE.transform.localScale = new Vector3(0, 0, 0);
					tmpLE.ignoreLayout = true;
				}
			}
		}
		
		// **************************
		// OVERRIDES LOADING
		// **************************	
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms);
			// Saving the original datas
			decalsDatas = jc;
		}
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy()
		{
			// Cleaning textures
			foreach (var srcTex in decalsSourceTextures)
			{
				Destroy(srcTex.Value);
			}
			decalsSourceTextures.Clear();
			
			foreach (var destTex in decalsDestTextures)
			{
				Destroy(destTex.Value);
			}
			decalsDestTextures.Clear();
			
			// Unloading assetbundle
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + SHADERS_BUNDLE_PATH);
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		// ***********************************************************
		// EXTERNAL TOOL - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** Chokapi Decal Maker START *********************
		public Texture2D GetResource(string path)
		{
			if (path == "" || path == null)
			{
				path = GetPackagePath(this) + "Custom/Scripts/VAMCCD/default.png";
			}
			Texture2D texture;
			byte[] tmppng = FileManagerSecure.ReadAllBytes(path);
			texture = new Texture2D(1, 1, TextureFormat.RGBA32, false);
			texture.LoadImage(tmppng);
			return texture;
		}
		// *********** Chokapi Decal Maker END *********************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		
		
		// *********** MacGruber_Utils.cs END *********************
	}
}